/* 
 * Copyright (C) 2003 by the libdvdnav project
 * 
 * This file is part of libdvdnav, a DVD navigation library.
 * 
 * libdvdnav is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * libdvdnav is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA
 *
 * $Id: menus.c,v 1.12 2003/03/29 14:19:01 mroi Exp $
 *
 */

#include <stdio.h>
#include <stdarg.h>
#include <getopt.h>
#include <unistd.h>
#include <inttypes.h>
#include <sys/ioctl.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <ansidecl.h>

#include "cvdvtypes.h"
#include "dvdnav.h"

#include "dvb_formats.h"

#include <linux/dvb/video.h>
#include <linux/dvb/audio.h>

/* which is the default language for menus/audio/subpictures? */
#define DVD_MENU_LANGUAGE "en"
#define DVD_AUDIO_LANGUAGE "en"
#define DVD_SUBPICTURE_LANGUAGE "nl"
#ifdef LANG_ja
# undef DVD_MENU_LANGUAGE
# undef DVD_AUDIO_LANGUAGE
# undef DVD_SUBPICTURE_LANGUAGE
# define DVD_MENU_LANGUAGE "ja"
# define DVD_AUDIO_LANGUAGE "en"
# define DVD_SUBPICTURE_LANGUAGE "ja"
#endif /* LANG_ja */

/* shall we use libdvdnav's read ahead cache? */
#define DVD_READ_CACHE 0

void DECCMD(int fd, struct decodercmd cmd)
{
	int err = ioctl(fd,_IO(CVDV_IOCTL_MAGIC,IOCTL_DECODER),&cmd);
	if (err) {
		perror("DECCMD: ioctl failed\n");
	}
}

#define  DO_IOCTL(fd, req, arg)\
	if (debug_level>1) printf("%d ioctl %d\n", __LINE__, req);\
	if (ioctl(fd, req, arg)!=0) perror("DO_IOCTL: ioctl failed")

#define  DO_IOCTL0(fd, req)\
	if (debug_level>1) printf("%d ioctl %d\n", __LINE__, req);\
	if (ioctl(fd, req)!=0) perror("DO_IOCTL0: ioctl failed")

#include <termios.h>

#include <X11/Xlib.h>
#include <X11/Xutil.h>
#include <X11/Xatom.h>
#define XK_MISCELLANY
#include <X11/keysymdef.h>
#include <X11/extensions/Xv.h>
#include <X11/extensions/Xvlib.h>

//#include <Xm/MwmUtil.h>

#include "EventNames.h"

#define GUID_YUV12_PLANAR 0x32315659

Display *dpy;
int xv_port = -1;
Window window, root;
GC gc;
Region zv_reg;
XRectangle zv_rect;

int subtitle_stream = 0;
int debug_level = 0;

static void help(void), help2(void), finish(int);

struct _sav_loc {
	int title;	//Title
	int part;	//Chapter
	int n_part; 	//Part
	FILE *loc;	//handle
	char *title_str; //Index?
} sav_loc;


void noborders(Display *display, Window win)
{
#if 0
	MwmHints mwmhints;
	Atom MOTIF_WM_HINTS;
	mwmhints.flags = MWM_HINTS_DECORATIONS;
	mwmhints.decorations = 0;
	MOTIF_WM_HINTS = XInternAtom(display, "_MOTIF_WM_HINTS", True);
	XChangeProperty(display, win, MOTIF_WM_HINTS, MOTIF_WM_HINTS,
			sizeof(long)*8, PropModeReplace,
			(unsigned char *) &mwmhints,
			sizeof(mwmhints) / sizeof(long) );
#endif
}

int zv_ratio_x = 16;
int zv_ratio_y = 9;

void zv_ratio_fixup(int *width, int *height, int *xoff, int *yoff)
{
	int h = *height;
	int w = *width;

	if ( (0==zv_ratio_x) ||
	     (0==zv_ratio_y) ) {
		return;
	}
	if ( w * zv_ratio_y < h * zv_ratio_x) {
		(*height) = ( (*width) * zv_ratio_y) / zv_ratio_x;
		if (yoff) {
			(*yoff) += (h-(*height) )/2;
		}
	} else if ( w * zv_ratio_y > h * zv_ratio_x) {
		(*width) = ( (*height) * zv_ratio_x) / zv_ratio_y;
		if (xoff) {
			(*xoff) += (w - (*width) )/2;
		}
	}
}

int _state;
#define STATE_NORMAL   0x00
#define STATE_TITLE    0x20
#define STATE_CHAPTER  0x30
#define STATE_MENU     0x40
#define STATE_SUBTITLE 0x50

int	get_state()
{
	return _state;
}

void set_state(int new_state)
{
	_state = new_state;
	switch(_state) {
	case STATE_NORMAL:
		// only key input
		XSelectInput(dpy, window, 
			     StructureNotifyMask | 
			     KeyPressMask    | KeyReleaseMask | ExposureMask
			     );
		break;
	case STATE_MENU:
		// enable mouse buttons and movements
		XSelectInput(dpy, window, 
			     StructureNotifyMask | PointerMotionMask | 
			     KeyPressMask    | KeyReleaseMask | 
			     ButtonPressMask | ButtonReleaseMask | ExposureMask
			     );
		break;
	}
}

int current_drawable_w, current_drawable_h;

void win_to_vid(int wx, int wy, int *vx, int *vy)
{
	(*vx) = (wx*720)/current_drawable_w;
	(*vy) = (wx*576)/current_drawable_h;
}

int	dbgprintf(int level, char *fmt, ...) 
{
	if (level >= debug_level) {
		va_list args;
		va_start(args, fmt);
		return vprintf(fmt, args);
	}
	return 0;
}

void zvwin(int permitted_df, int division)
{
	int vx, vy, vw, vh;
	int dx, dy, dw, dh;

	int x, y;
	unsigned int w, h, b, d;

	int le, to;
	int wi, he;

	char *p = "default";

	if (permitted_df==-1) {
		printf("zvwin: later...\n");
		return;
	}

	XGetGeometry(dpy, window, &root, &x, &y, &w, &h, &b, &d);

	/* get a correct drawable that fits in my window) */

	switch(permitted_df) {
	case 0:
	case 2: /* 16:9 */
		le = 8;
		to = 4;
		if (w*9 >= h*16) {	
			p = "16x9 width>height";
			he = (y+h) - to * 2;
			wi = ((he*16) / 9) + 25;
		} else {
			p = "16x9 height<width";
			wi = (x+w) - le * 2;
			he = ((wi*9) / 16) + 25;
		}

		//dx = 0;  dy = 0; dw = w;   dh = h;
		dx = le;  dy = to; dw = wi;   dh = he;
		vx = 0;  vy = 17; vw = 720; vh = 288;
		break;
	case 1: /* 4:3 */
		dx = 0;  dy = 0; dw = w;   dh = h;
		vx = 0;  vy = 17; vw = 640; vh = 288;
		p = "4x3 1";
		break;
	case 3: /* 4:3 */
		le = 0;
		to = 0;
		if (w*9 >= h*16) {	
			p = "16x9 width>height";
			he = (y+h) - to * 2;
			wi = (he*16)/9 ; //+25 ;
		} else {
			p = "16x9 height<width";
			wi = (x+w) - le * 2;
			he = (wi*9)/16 + 25;
		}


		//dx = 0;  dy = 0; dw = w;   dh = h;
		dx = le;  dy = to; dw = wi;   dh = he;
		vx = 0;   vy = 0;  vw = 720; vh = 288;
		break;
	default:
		printf("zvwin: %d should not happen!\n", permitted_df);
		return;
		break;
	}

	dbgprintf(1, "zvwin(%d \"%s\")\n", permitted_df, p);
	dbgprintf(1, "|x = %d y = %d w = %d h = %d b=%d d =%d\n",
	       x, y, w, h, b, d);

	dbgprintf(1, "| before: dw = %d dh = %d dx = %d dy = %d  ", dw, dh, dx, dy);
	dbgprintf(1,"|  before: vw = %d vh = %d vx = %d vy = %d\n", vw, vh, vx, vy);

	//zv_rect.x	= vx+100; zv_rect.y 	= vy+100;
	//zv_rect.width	= vw-100;zv_rect.height	= vh-100;
	XEmptyRegion(zv_reg);
	//XUnionRectWithRegion(&zv_rect, zv_reg, zv_reg);
	dbgprintf(1, "| zv_rect x = %d y = %d w = %d h = %d\n", zv_rect.x, zv_rect.y, zv_rect.width, zv_rect.height);
	//XSetRegion(dpy, gc, zv_reg);
	
	//vx = 0; vy = 0; vw = 720; vh = 576;
	//vx = 16;  vy = 20; vw = 720; vh = 625;
	//vw = 720;

	dbgprintf(1, "| after : dw = %d dh = %d dx = %d dy = %d  ", dw, dh, dx, dy);
	dbgprintf(1, "| after : vw = %d vh = %d vx = %d vy = %d\n", vw, vh, vx, vy);

	if (XvPutVideo(dpy, xv_port, window, gc, 
		       vx, vy, vw, vh, // video vx,vy vw,vh 
		       dx, dy, dw, dh  // drawable dx,dy dw,dh
		       ) != Success){
		printf("XvPutVideo failed.\n");
		//exit(-1);
	}

	current_drawable_h = dh;
	current_drawable_w = dw;

	dbgprintf(1, "|--------------------------------------------------\n");

	//	noborders(dpy, window);
	XFlush(dpy);
	
}
			       
void zvinit(int xs, int ys, int xe, int ye, int division)
{
	int screen;
	XVisualInfo vinfo;
	XSetWindowAttributes xswa;
	unsigned long mask;
	unsigned int p_version, p_release, p_request_base, p_event_base, p_error_base;
	unsigned int p_num_adaptors;
	XvAdaptorInfo *ai;
	int ret, x, y;
	unsigned int w, h, b, d;
	//  XEvent event;
	
	unsigned int nr_enq;
	XvEncodingInfo *xv_enq;
	
	int nr_fmt;
	XvImageFormatValues *xv_fmt;
	
	if ((dpy = XOpenDisplay(NULL)) == NULL){
		fprintf(stderr, "Cannot open Display.\n");
		return;
	}
	screen = DefaultScreen(dpy);
	if (!XMatchVisualInfo(dpy, screen, 16, TrueColor, &vinfo)
	    && !XMatchVisualInfo(dpy, screen, 15, TrueColor, &vinfo)){
		fprintf(stderr, "Cannot find the visual (15bpp or 16bpp).\n");
		exit(-1);
	}
	
	xswa.colormap =  XCreateColormap(dpy, DefaultRootWindow(dpy), vinfo.visual, AllocNone);
	xswa.event_mask = StructureNotifyMask | ExposureMask | PointerMotionMask | 
		KeyPressMask | ButtonPressMask;
	xswa.background_pixel = 0;
	xswa.border_pixel = 0;
	mask = CWBackPixel | CWBorderPixel | CWColormap | CWEventMask;
	
	window = XCreateWindow(dpy, DefaultRootWindow(dpy),
			       xs, ys, xe, ye, 0, vinfo.depth,
			       // 0, 0, 800, 480, 0, vinfo.depth,
			       //30, 0, 740, 440, 0, vinfo.depth,
			       //0, 15, 800, 465, 0, vinfo.depth,
			       InputOutput, vinfo.visual, mask, &xswa);
	
	gc = XCreateGC(dpy, window, 0, 0);

	zv_reg = XCreateRegion();	/* for testing purposes */

	XStoreName(dpy, window, "ZV");
	XSetIconName(dpy, window, "ZV");

	XMapWindow(dpy, window);
	
	XSync(dpy, 1);
	
	if ((ret = XvQueryExtension(dpy, &p_version, &p_release, &p_request_base, &p_event_base, &p_error_base)) != Success){
		fprintf(stderr, "Cannot find XV Extention.\n");
		exit(-1);
	}
	if ((ret = XvQueryAdaptors(dpy, DefaultRootWindow(dpy), &p_num_adaptors, &ai)) != Success){
		fprintf(stderr, "Cannot find XV adaptor.\n");
		exit(-1);
	}
	if (p_num_adaptors == 0){
		fprintf(stderr, "Cannot find XV adaptor.\n");
		exit(-1);
	}
	xv_port = ai[0].base_id;
	
	if ((ret = XvQueryEncodings(dpy, xv_port, &nr_enq, &xv_enq))  != Success) {
		fprintf(stderr, "Cannot Query Encodings\n");
		exit(-1);
	}
	
	if ((xv_fmt = XvListImageFormats(dpy, xv_port, &nr_fmt))  == NULL) {
		fprintf(stderr, "Cannot List Formats\n");
		exit(-1);
	}
	
	XGetGeometry(dpy, window, &root, &x, &y, &w, &h, &b, &d);
	printf("x = %d\ny = %d\nw = %d\nh = %d\nb=%d\nd =%d\n",
	       x, y, w, h, b, d);
	
}

char *dvdnav_event_types_1[] = {
	"DVDNAV_BLOCK_OK",
	"DVDNAV_NOP",
	"DVDNAV_STILL_FRAME",
	"DVDNAV_SPU_STREAM_CHANGE",
	"DVDNAV_AUDIO_STREAM_CHANGE",
	"DVDNAV_VTS_CHANGE",
	"DVDNAV_CELL_CHANGE",
	"DVDNAV_NAV_PACKET",
	"DVDNAV_STOP",
	"DVDNAV_HIGHLIGHT",
	"DVDNAV_SPU_CLUT_CHANGE",
	"unknown event",
	"DVDNAV_HOP_CHANNEL",
	"DVDNAV_WAIT"
};

char *dvdnav_event_types[] = {
	"BLOCK_OK           ",
	"NOP                ",
	"STILL_FRAME        ",
	"SPU_STREAM_CHANGE  ",
	"AUDIO_STREAM_CHANGE",
	"VTS_CHANGE         ",
	"CELL_CHANGE        ",
	"NAV_PACKET         ",
	"STOP               ",
	"HIGHLIGHT          ",
	"SPU_CLUT_CHANGE    ",
	"unknown event      ",
	"HOP_CHANNEL        ",
	"WAIT               "
};

char loop_string[] = "-/|\\";

void print_event(int event) 
{
	static int prev_event = -1;
	static int need_lf = 0;
	static int loop_var = 0;
	if (debug_level>1) {
		if ( (debug_level <=3) && 
		     (prev_event == event) ) {
			loop_var = loop_var>3 ? 0 : loop_var+1;
			printf("%c\b", loop_string[loop_var]);fflush(stdout);
			need_lf = 1;
		} else {
			if (need_lf == 1) {
				printf("\n");
				need_lf = 0;
			}
			loop_var = 0;
			if ( (event < DVDNAV_BLOCK_OK) || 
			     (event > DVDNAV_WAIT) ) {
				printf("unknown event\n");
			} else {
				printf("event (%2d) '%s' ", event, dvdnav_event_types[event]);
				need_lf = 1;
			}
			prev_event = event;
		}
	}
}



static int option_index = 0;
static struct option long_options[] = {
    {"audio", required_argument, NULL, 'a'},
    {"astream",required_argument, NULL, 'm'},
    {"subtitle",required_argument, NULL, 's'},
    {"help", no_argument , NULL, 'h'},
    {0, 0, 0, 0}
};

#if 0
char navipack[1024];
#endif

void set_subtitle(int fd, unsigned char subtitle_stream)
{
	video_spu_t Aspu;
	Aspu.active = 	 ((subtitle_stream & 0x80)==0x80) ? 0 : 1;
	Aspu.stream_id = (int) (subtitle_stream & 0x1f);
	printf("set_subtitle: subtitles %s: stream = %d\n", 
	       Aspu.active ? "On" : "Off", 
	       Aspu.stream_id);

	DO_IOCTL(fd, VIDEO_SET_SPU, &Aspu);
}

void highlight_button(dvdnav_t *dvdnav, int fd, int button, int active)
{
	pci_t *pci;
	dsi_t *dsi;
	btni_t *b;
	int i;
	unsigned char coli[4];
	video_highlight_t vh;	
	/* Applications with fifos should not use these functions to retrieve NAV packets,
	 * they should implement their own NAV handling, because the packet you get from these
	 * functions will already be ahead in the stream which can cause state inconsistencies.
	 * Applications with fifos should therefore pass the NAV packet through the fifo
	 * and decoding pipeline just like any other data. */
	if (debug_level>1) printf("button %d active %d\n",
				  button, active);
	pci = dvdnav_get_current_nav_pci(dvdnav);
	dsi = dvdnav_get_current_nav_dsi(dvdnav);
	b = &(pci->hli.btnit[button - 1]);
	i = b->btn_coln - 1;
	vh.active 	= active;

	vh.contrast1 	= coli[0] = pci->hli.btn_colit.btn_coli[i][active] >> 24;
	vh.contrast2 	= coli[1] = pci->hli.btn_colit.btn_coli[i][active] >> 16;
	vh.color2 	= coli[2] = pci->hli.btn_colit.btn_coli[i][active] >> 8;
	vh.color1 	= coli[3] = pci->hli.btn_colit.btn_coli[i][active];
	vh.xpos 	= vh.ypos = 0;

	vh.xpos = 
		( ( ( ( ( b->x_start & 0x3f0) >> 4) & 0x3f) | ( (b->btn_coln << 6)      ) ) << 16) +  // xpos[0]
		( ( ( ( ( b->x_start & 0xf  ) << 4) & 0xf0) | ( (b->x_end & 0x300) >> 8 ) ) << 8 ) +  // xpos[1]
		( ( ( ( ( b->x_end   & 0xff )     )       )                               )      );   // xpos[2]

	vh.ypos =
		( ( ( ( ( b->y_start & 0x3f0) >> 4) & 0x3f) | ( (b->auto_action_mode<<6)) ) << 16) +  // ypos[0]
		( ( ( ( ( b->y_start & 0xf  ) << 4) & 0xf0) | ( (b->y_end & 0x300) >> 8 ) ) << 8 ) +  // ypos[1]
		( ( ( ( ( b->y_end   & 0xff )     )       )                               )      );   // ypos[2]
	if (debug_level>1) {
		printf("coli[0] = %3d coli[1] = %3d coli[2] = %3d coli[3] = %3d "
		       "auto_action_mode = %d btn_coln = %d "
		       "x_start = %3d x_end = %3d y_start = %3d y_end = %3d\n",
		       coli[0], coli[1], coli[2], coli[3], 
		       b->auto_action_mode, b->btn_coln, 
		       b->x_start, b->x_end,
		       b->y_start, b->y_end);
		printf("vh.contrast1 = %3d vh.contrast2 = %3d vh color1    = %3d vh.color2    = %3d "
		       "vh.xpos = %8x vh.ypos = %8x\n",
		       vh.contrast1, vh.contrast2, vh.color1, vh.color2, 
		       vh.xpos, vh.ypos);
	}
	DO_IOCTL(fd, VIDEO_SET_HIGHLIGHT, &vh);
	//set_subtitle(fd, 1, subtitle_stream);
}


int div_tab_x[] = {800, 720, 640, 560, 480, 400, 320, 240, 160, 80};
int div_tab_y[] = {450, 405, 360, 315, 270, 225, 180, 135, 90, 45};

int main(int argc, char **argv)
{
    int i, c;
    int written;
    int len_w;
    char *audio_type = "AC3";
    int audio_stream = 0;
    int division = 1;
    char *path = NULL;
    dvdnav_t *dvdnav;
    uint8_t mem[DVD_VIDEO_LB_LEN];
    int finished = 0;
    int fd;
    int fda;
    int skip_wait = 0;
    int wait_time = 0;
    int wait_time_elapsed = 0;
    int still_frame = 0;
    int skip_still_frame = 0;
   
    //struct decodercmd decmd;

    ///video_spu_t	Aspu;
    video_spu_palette_t Aspupalette;
    //video_highlight_t hl;

    unsigned char clut[3*16];
    int button;
    int menu_dir;
#define MENU_LEFT  1
#define MENU_DOWN  2
#define MENU_UP    3
#define MENU_RIGHT 4
#define MENU_ENTER 5
    fd_set wfds;
    struct timeval t;
    int stopped;
    int hop;
    int8_t curr_perm_scale = -1;/* wait for init */

    while ((c = getopt_long(argc, argv, "D:d:a:m:s:h", long_options, &option_index)) >= 0){
	switch (c){
	case 'a':
	    audio_type = optarg;
	    break;
	case 'd':
		division = atoi(optarg);
		if ((division < 0) || (division>9)) {
			fprintf(stderr, "division must be 0 <= division <= 9\n");
			return(1);
		}
		break;
	case 'm':
		audio_stream = atoi(optarg);
		break;
	case 's':
		subtitle_stream = atoi(optarg);
		break;
	case 'D':
		debug_level = atoi(optarg);
		break;
	case 'h':
	case '?':
	default:
	    help();
	    exit(0);
	}
    }

    if (optind < argc){
	path = argv[optind];
    } else {
	help();
	exit(-1);
    }


    if (open_av(&fd, &fda, 0) < 0) exit(1);
    if (audioSetStreamtype(fda, audio_type) < 0) exit(1);
    if (videoSetStreamtype(fd, "PROG") < 0) exit(1);

    if (fda) audioSetAVSync(fda, 1);

    videoSelectSource(fd, VIDEO_SOURCE_MEMORY);
    //    videoPlay(fd);

	 
    //    if ((fd = open(DEVICE_NAME, O_RDWR|O_NONBLOCK)) < 0){
    //	perror(DEVICE_NAME);
    //	exit(-1);
    //}

    zvinit(0, 150, div_tab_x[division], div_tab_y[division], division);
    
    //decmd.cmd = Decoder_Set_Streamtype; decmd.param1 = stream_DVD; DECCMD(fd, decmd);
    //DO_IOCTL(fd, VIDEO_SET_STREAMTYPE, VIDEO_CAP_SPU);

    //set_subtitle(fd, 1, subtitle_stream);

/*     decmd.cmd = Decoder_SPU; */
/*     decmd.param1 = (subtitle_stream < 0) ? 32 : subtitle_stream; */
/*     decmd.param2 = 1; */
/*     DECCMD(fd, decmd); */

    /* open dvdnav handle */
    if (dvdnav_open(&dvdnav, path) != DVDNAV_STATUS_OK) {
	fprintf(stderr, "Error: dvdnav_open %s\n", path);
	exit(1);
    }
#if 0
    dvdnav_set_region_mask(dvdnav, 0x02);
#endif

    {
	    const char *title_str = NULL;
	    dvdnav_get_title_string(dvdnav, &title_str);
	    printf("Title name: %s\n", title_str ? title_str : "?! no title");
    }

    /* set read ahead cache usage */
    if (dvdnav_set_readahead_flag(dvdnav, DVD_READ_CACHE) != DVDNAV_STATUS_OK) {
	fprintf(stderr, "Error: dvdnav_set_readahead_flag: %s\n", dvdnav_err_to_string(dvdnav));
	exit(2);
    }
  
    /* set the language */
    if (dvdnav_menu_language_select(dvdnav, DVD_MENU_LANGUAGE) != DVDNAV_STATUS_OK ||
	dvdnav_audio_language_select(dvdnav, DVD_AUDIO_LANGUAGE) != DVDNAV_STATUS_OK ||
	dvdnav_spu_language_select(dvdnav, DVD_SUBPICTURE_LANGUAGE) != DVDNAV_STATUS_OK) {
	fprintf(stderr, "Error: setting languages: %s\n", dvdnav_err_to_string(dvdnav));
	exit(2);
    }
  
    /* set the PGC positioning flag to have position information relatively to the
     * whole feature instead of just relatively to the current chapter */
    if (dvdnav_set_PGC_positioning_flag(dvdnav, 1) != DVDNAV_STATUS_OK) {
	fprintf(stderr, "Error: dvdnav_set_PGC_positioning_flag: %s\n", dvdnav_err_to_string(dvdnav));
	exit(2);
    }

    stopped = 0;
    hop = 0;

    /* the read loop which regularly calls dvdnav_get_next_block
     * and handles the returned events */
    while (!finished) {
	    int result, event, len;
	    uint8_t *buf = mem;
	    int nr_events;
	    XEvent	x_ev;
	    char	x_inp[4];
	    int	x_inp_size=4;
	    KeySym	keysym;
	    XComposeStatus compose;
	    int	charcount;
	    int	x_input_valid = 0;
	    
	    t.tv_sec = wait_time <= 0 ? 0 : wait_time ;
	    t.tv_usec = 0;
	    if ((nr_events=XPending(dpy))!=0) {
		    //printf("We have Xevents %d\n", nr_events);
		    while(nr_events--) {
			    XNextEvent(dpy, &x_ev);
			    switch(x_ev.type) {
			    default:
				    if (debug_level>0) printf("X event  %s\n", eventNames[x_ev.type]);
				    break;
			    case Expose:
				    if (debug_level>0) printf("Expose %d\n", x_ev.xexpose.count);
				    if (x_ev.xexpose.count == 0) {
					    zvwin(curr_perm_scale, division);
				    }
				    break;
			    case ResizeRequest:
				    if (debug_level>0) printf("Resize w:%d h:%d\n", x_ev.xresizerequest.width,
							      x_ev.xresizerequest.height);
				    break;
			    case ConfigureNotify:
				    dbgprintf(1, "Configure:%d h:%d\n", 
					      x_ev.xconfigure.width,
					      x_ev.xconfigure.height);

				    break;
			    case KeyPress:
				    charcount = XLookupString( (XKeyEvent*)&x_ev, x_inp, x_inp_size, &keysym, &compose);
				    x_input_valid = 1;
				    if (debug_level>0) printf("KeyPress: count %d, (keysym: %u08x)\n", charcount, 
							      (unsigned int)keysym);
				    break;
			    case MotionNotify:
				    if (get_state() == STATE_MENU) {
					    int	v_x, v_y;
					    pci_t *pci =dvdnav_get_current_nav_pci(dvdnav);
					    //set_state(STATE_NORMAL);
					    //v_x = (x_ev.xmotion.x*720)/div_tab_x[division];
					    //v_y = (x_ev.xmotion.y*576)/div_tab_y[division];
					    win_to_vid(x_ev.xmotion.x, 
						       x_ev.xmotion.y,
						       &v_x,
						       &v_y);

					    if (debug_level>0) printf("Mouse: x: %d y: %d v_x: %d v_y:%d \n", 
								      x_ev.xmotion.x, x_ev.xmotion.y, v_x, v_y);
					    if (pci) {
						    dvdnav_mouse_select(dvdnav, pci, v_x, v_y);
					    }
				    }
				    break;
			    case ButtonPress:
				    if (get_state() == STATE_MENU) {
					    int	v_x, v_y;
					    pci_t *pci =dvdnav_get_current_nav_pci(dvdnav);
					    win_to_vid(x_ev.xbutton.x, 
						       x_ev.xbutton.y,
						       &v_x,
						       &v_y);
					    //v_x = (x_ev.xbutton.x*720)/div_tab_x[division];
					    //v_y = (x_ev.xbutton.y*576)/div_tab_y[division];
					    if (debug_level>0) printf("Button: x: %d y: %d v_x: %d v_y:%d \n", 
								      x_ev.xmotion.x, x_ev.xmotion.y, v_x, v_y);
					    if (pci) {
						    dvdnav_mouse_activate(dvdnav, pci, v_x, v_y);
						    menu_dir = 0;
						    
						    wait_time_elapsed = 0;
						    skip_wait = 1;
						    wait_time = 0;
					    }
				    }
				    break;
			    }
		    }
		    /*
		     * handle keyboard events
		     */
		    if (x_input_valid == 1) {
			    c = keysym;
			    x_input_valid = 0;
			    
			    if (debug_level >=1) printf("key press %c\n", c);
			    switch (c){
			    case 'w':
#if 1
				    /* 			    decmd.cmd = Decoder_AVSYNC; */
				    /* 			    decmd.param1 = 2; */
				    /* 			    decmd.param2 = 30000; */
				    /* 			    DECCMD(fd, decmd); */
				    DO_IOCTL(fda, AUDIO_SET_AV_SYNC, 1);
#endif
				    break;
			    case '?':
				    help2();
				    break;
			    case 'i': /* info */
			    {
				    int n_title, title, part, n_part;
				    dvdnav_current_title_info(dvdnav, &title, &part);
				    dvdnav_get_number_of_parts(dvdnav, title, &n_part);
				    printf("[Current] Title: %2d Chapter: %2d Part:%d\n", 
					   title, part, n_part);
				    dvdnav_get_number_of_titles(dvdnav, &n_title);
				    for (title = 1; title <= n_title; title++){
					    dvdnav_get_number_of_parts(dvdnav, title, &n_part);
					    printf("  Title %2d: %2d chapters\n", 
						   title, n_part);
				    }
			    }
			    break;
			    case ' ': /* pause toggle */
				    //stopped = !stopped;
				    //if (stopped) {
				    //	    printf("play\n");
				    //DO_IOCTL0(fd, VIDEO_CONTINUE);
				    //} else {
				    //printf("pauze\n");
				    //    //DO_IOCTL0(fd, VIDEO_FREEZE);
				    //}
				    break;
			    case 't':
				    printf("title menu\n");
				    if (dvdnav_menu_call(dvdnav, DVD_MENU_Title) == DVDNAV_STATUS_OK)
					    hop = 1;
				    set_state(STATE_NORMAL);
				    set_subtitle(fd, 0x80);
				    break;
			    case 'r':
				    printf("root menu\n");
				    if (dvdnav_menu_call(dvdnav, DVD_MENU_Root) == DVDNAV_STATUS_OK)
					    hop = 1;
				    set_state(STATE_NORMAL);
				    set_subtitle(fd, 0x80);
				    break;
			    case 'n':
			    case XK_Page_Down:
				    dvdnav_next_pg_search(dvdnav);
				    break;
			    case 'p':
			    case XK_Page_Up:
				    dvdnav_prev_pg_search(dvdnav);
				    break;
			    case 'T': /* title */
				    // if (get_state() == STATE_NORMAL){
				    //    set_state(STATE_TITLE|STATE_NUMINPUT);
				    //    keyinp = 0;
				    //}
				    break;
			    case 'C': /* chapter */
				    //if (get_state() == STATE_NORMAL){
				    //    set_state(STATE_CHAPTER|STATE_NUMINPUT);
				    //    keyinp = 0;
				    // }
				    break;
			    case 'a': /* select audio track */
				    break;
			    case 's': /* select subtitle stream */
				    break;
			    case 'h':
			    case XK_Left:
			    {
				    pci_t *pci =dvdnav_get_current_nav_pci(dvdnav);
				    menu_dir = MENU_LEFT;
				    if (get_state() != STATE_MENU){
					    dvdnav_sector_search(dvdnav, -5000, SEEK_CUR);
				    }
				    //if (skip_wait==0) skip_wait = 1;
				    if (still_frame==1) dvdnav_left_button_select(dvdnav, pci);
			    }
			    break;
			    case 'j':
			    case XK_Down:
			    {
				    pci_t *pci =dvdnav_get_current_nav_pci(dvdnav);
				    menu_dir = MENU_DOWN;
				    if (get_state() != STATE_MENU){
					    dvdnav_sector_search(dvdnav, 15000, SEEK_CUR);
				    }
				    //if (skip_wait==0) skip_wait = 1;
				    if (still_frame==1) dvdnav_lower_button_select(dvdnav, pci);
			    }
			    break;
			    case 'k':
			    case XK_Up:
			    {
				    pci_t *pci =dvdnav_get_current_nav_pci(dvdnav);
				    menu_dir = MENU_UP;
				    if (get_state() != STATE_MENU){
					    dvdnav_sector_search(dvdnav, -15000, SEEK_CUR);
				    }
				    //if (skip_wait==0) skip_wait = 1;
				    if (still_frame==1) dvdnav_upper_button_select(dvdnav, pci);
			    }
			    break;
			    case 'l':
			    case XK_Right:
			    {
				    pci_t *pci =dvdnav_get_current_nav_pci(dvdnav);
				    menu_dir = MENU_RIGHT;
				    if (get_state() != STATE_MENU){
					    dvdnav_sector_search(dvdnav, 5000, SEEK_CUR);
				    }
				    //if (skip_wait==0) skip_wait = 1;
				    if (still_frame==1) dvdnav_right_button_select(dvdnav, pci);
			    }
			    break;
			    case XK_Return:
				    menu_dir = MENU_ENTER;
				    wait_time_elapsed = 1;
				    skip_wait = 1;
				    if (still_frame==1) skip_still_frame = 1;
				    break;
			    case 'q':
				    printf("quit\n");
				    finished = 1;
				    int title, part, n_part;
				    dvdnav_current_title_info(dvdnav, &title, &part);
				    dvdnav_get_number_of_parts(dvdnav, title, &n_part);
				    printf("[Current] Title: %2d Chapter: %2d n_part %d\n", 
					   title, part, n_part);
				    break;
				    break;
				    break;
			    }
		    } while (stopped  &&  !finished);
		    /* timeout expired */
		    if (wait_time>0) {
			    wait_time_elapsed = 1;
		    }
	    }
	    if (finished)
		    continue;
	    
#if 0
	    decmd.cmd = Decoder_GetNavi;
	    decmd.data1 = navipack;
	    if ((result = DECCMD(fd, decmd)) < 0){
		    fprintf(stderr, "NAVIPACK Fail\n");
	    } else if (result > 0){
		    fprintf(stderr, "NAVIPACK %d\n", result);
	    }
#endif
	    
	    /* the main reading function */
#if DVD_READ_CACHE
	    result = dvdnav_get_next_cache_block(dvdnav, &buf, &event, &len);
#else
	    result = dvdnav_get_next_block(dvdnav, buf, &event, &len);
#endif
	    if (result == DVDNAV_STATUS_ERR) {
		    fprintf(stderr, "Error: getting next block: %s\n", dvdnav_err_to_string(dvdnav));
		    finish(3);
	    }
	    
	    print_event(event);
	    
	    switch (event){
	    case DVDNAV_BLOCK_OK:
		    if (1 ) {
			    written = 0;
			    while(len > 0 && written<len ) {
				    FD_ZERO(&wfds);
				    FD_SET(fd, &wfds);
				    t.tv_sec = 0;
				    t.tv_usec = 500000;
				    if (select(fd+1, NULL, &wfds, NULL, &t) == 1) {
					    len_w = write(fd, buf+written, len-written);
					    if (len_w > 0) written += len_w;
				    } else {
					    perror("select time out");
					    fprintf(stderr, 
						    "len     %d\n"
						    "len_w   %d\n"
						    "written %d\n",
						    len, len_w, written);
					    break;
				    }
			    }
		    }
		    break;
	    case DVDNAV_NOP:
		    /* Nothing to do here. */
		    break;
	    case DVDNAV_STILL_FRAME: 
		    /* We have reached a still frame. A real player application would wait
		     * the amount of time specified by the still's length while still handling
		     * user input to make menus and other interactive stills work.
		     * A length of 0xff means an indefinite still which has to be skipped
		     * indirectly by some user interaction. */
		    {
			    dvdnav_still_event_t *still_event = (dvdnav_still_event_t *)buf;
			    if (still_frame) { // was here already
				    if ((wait_time_elapsed) || 
					(skip_still_frame)) { 
					    printf("Skipping indefinite length still frame\n");
					    skip_still_frame = 0;
					    still_frame = 0;
					    dvdnav_still_skip(dvdnav);
				    }
			    } else {
				    if (still_event->length < 0xff) {
					    printf("Skipping %d seconds of still frame\n", 
						   still_event->length);
					    wait_time = still_event->length;
				    } else {
					    wait_time = -1;
				    }
				    still_frame = 1;
				    wait_time_elapsed = 0;
			    }
		    }
		    break;
	    case DVDNAV_WAIT:
		    /* We have reached a point in DVD playback, where timing is critical.
		     * Player application with internal fifos can introduce state
		     * inconsistencies, because libdvdnav is always the fifo's length
		     * ahead in the stream compared to what the application sees.
		     * Such applications should wait until their fifos are empty
		     * when they receive this type of event. */
		    sleep(1);
		    
		    //if (skip_wait) {
		    printf("Skipping wait condition\n");
		    dvdnav_wait_skip(dvdnav);
		    skip_wait = 0;
		    //}
		    break;
	    case DVDNAV_SPU_CLUT_CHANGE:
		    /* Player applications should pass the new colour lookup table to their
		     * SPU decoder */
		    for (i = 0; i < len/4; i++){
			    
			    clut[i*3+0] = buf[i*4+2];
			    clut[i*3+1] = buf[i*4+1];
			    clut[i*3+2] = buf[i*4+0];
		    }
		    
		    Aspupalette.palette = clut;
		    Aspupalette.length  = len / 4 * 3;
		    DO_IOCTL(fd, VIDEO_SET_SPU_PALETTE, &Aspupalette);
		    break;
	    case DVDNAV_SPU_STREAM_CHANGE:
		    /* Player applications should inform their SPU decoder to switch channels */
		    {
			    dvdnav_spu_stream_change_event_t *spu_ev;
			    
			    spu_ev = (dvdnav_spu_stream_change_event_t *)buf;
			    printf("SPU Stream Change: wide %02x letterbox %02x pan&scan %02x %06x\n",
				   spu_ev->physical_wide, spu_ev->physical_letterbox,
				   spu_ev->physical_pan_scan, spu_ev->logical);
			    
			    set_subtitle(fd, spu_ev->physical_wide);
		    }
		    break;
	    case DVDNAV_AUDIO_STREAM_CHANGE:
		    /* Player; applications should inform their audio decoder to switch channels */
		    {
			    dvdnav_audio_stream_change_event_t *audio_ev;
			    
			    audio_ev = (dvdnav_audio_stream_change_event_t *)buf;
			    printf("AUDIO Stream Change: %d %d\n", audio_ev->physical, audio_ev->logical);
			    if (audio_ev->physical >= 0  &&  audio_ev->physical != audio_stream){
				    DO_IOCTL(fda, AUDIO_SET_ID, audio_stream);
			    }
		    }
		    break;
	    case DVDNAV_HIGHLIGHT:
		    /* Player applications should inform their overlay engine to highlight the 
		     * given button */
		    {
			    dvdnav_highlight_event_t *highlight_event = (dvdnav_highlight_event_t *)buf;
			    printf("Selected button %d\n", highlight_event->buttonN);
			    button = highlight_event->buttonN;
			    highlight_button(dvdnav, fd, button, 1);
		    }
		    break;
	    case DVDNAV_VTS_CHANGE:
		    /* Some status information like video aspect and video scale permissions do
		     * not change inside a VTS. Therefore this event can be used to query such
		     * information only when necessary and update the decoding/displaying
		     * accordingly. */
		    {
			    dvdnav_vts_change_event_t *vts_ev;
			    //video_attr_t v;
			    //unsigned short attr;
			    //unsigned int flush =1;
			    
			    vts_ev = (dvdnav_vts_change_event_t *)buf;
			    printf("VTS Change: %d->%d\n", vts_ev->old_vtsN, vts_ev->new_vtsN);
			    if (vts_ev->new_vtsN < 0)
				    break;
			    
			    curr_perm_scale = dvdnav_get_video_aspect(dvdnav);

			    //printf("Video attibutes %08x\n", attr);
			    //DO_IOCTL(fd, VIDEO_SET_ATTRIBUTES, attr);
			    zvwin(curr_perm_scale, division);
			    DO_IOCTL0(fd, VIDEO_CONTINUE);
		    }
		    break;
	    case DVDNAV_CELL_CHANGE:
		    /* Some status information like the current Title and Part numbers do not
		     * change inside a cell. Therefore this event can be used to query such
		     * information only when necessary and update the decoding/displaying
		     * accordingly. */
		    {
			    int tt = 0, ptt = 0;
			    unsigned int pos, len;
			    
			    dvdnav_current_title_info(dvdnav, &tt, &ptt);
			    dvdnav_get_position(dvdnav, &pos, &len);
			    printf("Cell change: Title %d, Chapter %d\n", tt, ptt);
			    printf("At position %.0f%% inside the feature (pos = %d)\n", 100 * (double)pos / (double)len, pos);
		    }
		    break;
	    case DVDNAV_NAV_PACKET:
		    /* A NAV packet provides PTS discontinuity information, angle linking information and
		     * button definitions for DVD menus. Angles are handled completely inside libdvdnav.
		     * For the menus to work, the NAV packet information has to be passed to the overlay
		     * engine of the player so that it knows the dimensions of the button areas. */
		    {
			    pci_t *pci;
			    dsi_t *dsi;
			    btni_t *b;
			    // unsigned char coli[4], posi[6];
			    video_highlight_t vh;	
			    /* Applications with fifos should not use these functions to retrieve NAV packets,
			     * they should implement their own NAV handling, because the packet you get from these
			     * functions will already be ahead in the stream which can cause state inconsistencies.
			     * Applications with fifos should therefore pass the NAV packet through the fifo
			     * and decoding pipeline just like any other data. */
			    pci = dvdnav_get_current_nav_pci(dvdnav);
			    dsi = dvdnav_get_current_nav_dsi(dvdnav);
			    
			    if(pci->hli.hl_gi.btn_ns > 0) {
				    
				    if (get_state() == STATE_NORMAL){
					    if (debug_level>0) {
						    printf("Found %i DVD menu buttons...\n", 
							   pci->hli.hl_gi.btn_ns);
						    for (i = 0; i < pci->hli.hl_gi.btn_ns; i++){
							    b = &(pci->hli.btnit[i]);
							    printf("Button %i top-left @ (%i,%i), "
								   " bottom-right @ (%i,%i)\n", 
								   i+1, b->x_start, b->y_start, 
								   b->x_end, b->y_end);
						    }
					    }
					    set_state(STATE_MENU);
					    menu_dir = 0;
				    }
				    /* validate button value */
				    if (button <= 0)
					    button = 1;
				    else if (button > pci->hli.hl_gi.btn_ns)
					    button = pci->hli.hl_gi.btn_ns;
				    b = &(pci->hli.btnit[button - 1]);
				    switch (menu_dir){
				    case MENU_LEFT:
					    button = b->left;
					    menu_dir = 0;
					    break;
				    case MENU_DOWN:
					    button = b->down;
					    menu_dir = 0;
					    break;
				    case MENU_UP:
					    button = b->up;
					    menu_dir = 0;
					    break;
				    case MENU_RIGHT:
					    button = b->right;
					    menu_dir = 0;
					    break;
				    }
				    
				    highlight_button(dvdnav, fd, button, 1);
				    
				    if (menu_dir == MENU_ENTER){
					    set_state(STATE_NORMAL);
					    menu_dir = 0;
					    vh.active = 0;
					    highlight_button(dvdnav, fd, button, 0);
					    
					    printf("Selecting button %i...\n", button);
					    /* This is the point where applications with fifos have to 
					     *  hand in a NAV packet
					     * which has traveled through the fifos. See the notes above. 
					     */
					    dvdnav_button_select_and_activate(dvdnav, pci, button);
					    //sleep(2);
					    skip_wait= 0;
					    wait_time_elapsed = 0;
					    //skip_still_frame = 1;
					    wait_time = 0;
					    hop = 2;
				    }
			    }
		    }
		    break;
	    case DVDNAV_HOP_CHANNEL:
		    /* This event is issued whenever a non-seamless operation has been executed.
		     * Applications with fifos should drop the fifos content to speed up responsiveness. */
		    printf("Hop Channel: %d\n", hop);
		    if (hop == 1){
			    DO_IOCTL(fd, VIDEO_CLEAR_BUFFER, 1); /* discard */
		    } else {
			    DO_IOCTL(fd, VIDEO_CLEAR_BUFFER, 1);
		    }
		    hop = 0;
		    break;
	    case DVDNAV_STOP:
		    /* Playback should end here. */
		    {
			    finished = 1;
		    }
		    break;
	    default:
		    printf("Unknown event (%i)\n", event);
		    finished = 1;
		    break;
	    }
#if DVD_READ_CACHE
	    dvdnav_free_cache_block(dvdnav, buf);
#endif
    }
    
    /* destroy dvdnav handle */
    if (dvdnav_close(dvdnav) != DVDNAV_STATUS_OK) {
	    fprintf(stderr, "Error: dvdnav_close: %s\n", dvdnav_err_to_string(dvdnav));
	    finish(5);
    }
    close(fd);
    
    finish(0);
    return 0;
}

static void finish(int code)
{
	if (dpy) XCloseDisplay(dpy);
	exit(code);
}

static void help(void)
{
    fprintf (stderr, "Usage: margi-dvd [options] {device|directory}\n");
    fprintf (stderr, "options:\n");
    fprintf (stderr, "  --help,    -h:  print help message\n");  
    fprintf (stderr, "  --audio,   -a:  set audio type (AC3, LPCM, MPEG)\n");
    fprintf (stderr, "  --subtitle,-s:  set subtitle stream number (default:-1)\n");
}

static void help2(void)
{
    printf("?           : print this message\n");
    printf("q           : exit\n");
    printf("i           : print title information\n");
    printf("t           : title menu\n");
    printf("r           : root menu\n");
    printf("aN          : audio stream N\n");
    printf("sN          : subtitle stream N\n");
    printf("TN          : title N\n");
    printf("CN          : chapter N\n");
    printf("n           : next chapter\n");
    printf("p           : prev chapter\n");
    printf("SPACE       : pause (toggle)\n");
    printf("CURSOR left : skip backword, menu select left\n");
    printf("CURSOR down : skip backword, menu select down\n");
    printf("CURSOR up   : skip forward,  menu select up\n");
    printf("CURSOR right: skip forward,  menu select right\n");
    printf("h           : menu select left\n");
    printf("j           : menu select down\n");
    printf("k           : menu select up\n");
    printf("l           : menu select right\n");
}
